
import logging
import json

from aiohttp import web
from aiohttp import BasicAuth

clogger = logging.getLogger("dnstap_receiver.console")

class Handlers:
    def __init__(self, apikey, basicauth, stats, cfg_stats):
        self.api_key = apikey
        self.basic_auth = basicauth
        self.cfg_stats = cfg_stats
        self.stats = stats
        self.top = cfg_stats["max-items"]
        
    def check_auth(self, request):
        """check api key value"""
        # support basic auth or x-api-key authentication
        req_auth = request.headers.get('X-API-Key')
        basic_auth = request.headers.get('Authorization')

        if req_auth is None and basic_auth is None:
            return False
        
        if req_auth is not None:
            if req_auth != self.api_key:
                return False
                
        if basic_auth is not None:
            auth = BasicAuth(login="").decode(auth_header=basic_auth)

            if self.basic_auth != auth:
                return False
                
        return True
   
    async def handle_reset(self, request):
        auth = self.check_auth(request=request)
        if not auth:
            return web.Response(status=401)
      
        self.stats.reset()
        return web.Response(status=204)
    
    async def handle_metrics(self, request):
        auth = self.check_auth(request=request)
        if not auth:
            return web.Response(status=401)

        filters = ["pps", "domains", "clients", "query/bytes", "response/bytes",
                   "qps", "query", "query/udp", "query/tcp", "query/inet", "query/inet6", 
                   "query/a", "query/aaaa", "query/svr",
                   "rps", "response", "response/udp", "response/tcp", "response/inet", "response/inet6", 
                   "response/nxdomain", "response/noerror", "response/servfail", "response/refused",
                   "response/latency0_1", "response/latency1_10", "response/latency10_50",
                   "response/latency50_100", "response/latency100_1000", "response/latency_slow"]
                   
        # global counters
        counters = self.stats.get_counters(stream=None, filters=filters)
  
        p = []
        
        p.append( "# HELP dnstap_queries_bytes_total Total number of bytes for queries" )
        p.append( "# TYPE dnstap_queries_bytes_total counter" )
        p.append( "dnstap_queries_bytes_total %s" % counters["query/bytes"] )
        
        p.append( "# HELP dnstap_responses_bytes_total Total number of bytes for responses" )
        p.append( "# TYPE dnstap_responses_bytes_total counter" )
        p.append( "dnstap_responses_bytes_total %s" % counters["response/bytes"] )
        
        p.append( "# HELP dnstap_pps_total Number of packets per second received" )
        p.append( "# TYPE dnstap_pps_total gauge" )
        p.append( "dnstap_pps %s" % counters["pps"] )
        
        p.append( "# HELP dnstap_qps_total Number of queries per second received" )
        p.append( "# TYPE dnstap_qps_total gauge" )
        p.append( "dnstap_qps %s" % counters["qps"] )
        
        p.append( "# HELP dnstap_rps_total Number of responses per second received" )
        p.append( "# TYPE dnstap_rps_total gauge" )
        p.append( "dnstap_rps %s" % counters["rps"] )
        
        p.append( "# HELP dnstap_domains_total Number of domains asked" )
        p.append( "# TYPE dnstap_domains_total counter" )
        p.append( "dnstap_domains %s" % counters["domains"] )
        
        p.append( "# HELP dnstap_clients_total Number of clients asked" )
        p.append( "# TYPE dnstap_clients_total counter" )
        p.append( "dnstap_clients %s" % counters["clients"] )
        
        p.append( "# HELP dnstap_queries_total Number of queries received" )
        p.append( "# TYPE dnstap_queries_total counter" )
        p.append( "dnstap_queries %s" % counters["query"] )
        
        p.append( "# HELP dnstap_queries_udp_total Number of UDP queries received" )
        p.append( "# TYPE dnstap_queries_udp_total counter" )
        p.append( "dnstap_queries_udp %s" % counters["query/udp"] )
        
        p.append( "# HELP dnstap_queries_tcp_total Number of TCP queries received" )
        p.append( "# TYPE dnstap_queries_tcp_total counter" )
        p.append( "dnstap_queries_tcp %s" % counters["query/tcp"] )
        
        p.append( "# HELP dnstap_queries_inet_total Number of IPv4 queries received" )
        p.append( "# TYPE dnstap_queries_inet_total counter" )
        p.append( "dnstap_queries_inet %s" % counters["query/inet"] )
        
        p.append( "# HELP dnstap_queries_inet6_total Number of IPv6 queries received" )
        p.append( "# TYPE dnstap_queries_inet6_total counter" )
        p.append( "dnstap_queries_inet6 %s" % counters["query/inet6"] )
        
        p.append( "# HELP dnstap_queries_a_total Number of A queries received" )
        p.append( "# TYPE dnstap_queries_a_total counter" )
        p.append( "dnstap_queries_a %s" % counters["query/a"] )
        
        p.append( "# HELP dnstap_queries_aaaa_total Number of AAAA queries received" )
        p.append( "# TYPE dnstap_queries_aaaa_total counter" )
        p.append( "dnstap_queries_aaaa %s" % counters["query/aaaa"] )
        
        p.append( "# HELP dnstap_queries_svr_total Number of SVR queries received" )
        p.append( "# TYPE dnstap_queries_svr_total counter" )
        p.append( "dnstap_queries_svr %s" % counters["query/svr"] )
        
        p.append( "# HELP dnstap_responses_total Number of responses received" )
        p.append( "# TYPE dnstap_responses_total counter" )
        p.append( "dnstap_responses %s" % counters["response"] )
        
        p.append( "# HELP dnstap_response_udp_total Number of UDP responses received" )
        p.append( "# TYPE dnstap_response_udp_total counter" )
        p.append( "dnstap_response_udp %s" % counters["response/udp"] )
        
        p.append( "# HELP dnstap_responses_tcp_total Number of TCP responses received" )
        p.append( "# TYPE dnstap_response_tcp_total counter" )
        p.append( "dnstap_response_tcp %s" % counters["response/tcp"] )
        
        p.append( "# HELP dnstap_response_inet_total Number of IPv4 responses received" )
        p.append( "# TYPE dnstap_response_inet_total counter" )
        p.append( "dnstap_response_inet %s" % counters["response/inet"] )
        
        p.append( "# HELP dnstap_response_inet6_total Number of IPv6 responses received" )
        p.append( "# TYPE dnstap_response_inet6_total counter" )
        p.append( "dnstap_response_inet6 %s" % counters["response/inet6"] )
        
        p.append( "# HELP dnstap_responses_noerror_total Number of NOERROR answers" )
        p.append( "# TYPE dnstap_responses_noerror_total counter" )
        p.append( "dnstap_responses_noerror %s" % counters["response/noerror"] )
        
        p.append( "# HELP dnstap_responses_nxdomain_total Number of NXDOMAIN answers" )
        p.append( "# TYPE dnstap_responses_nxdomain_total counter" )
        p.append( "dnstap_responses_nxdomain %s" % counters["response/nxdomain"] )
        
        p.append( "# HELP dnstap_responses_servfail_total Number of SERVFAIL answers" )
        p.append( "# TYPE dnstap_responses_servfail_total counter" )
        p.append( "dnstap_responses_servfail %s" % counters["response/servfail"] )

        p.append( "# HELP dnstap_responses_refused_total Number of REFUSED answers" )
        p.append( "# TYPE dnstap_responses_refused_total counter" )
        p.append( "dnstap_responses_refused %s" % counters["response/refused"] )
        
        p.append( "# HELP dnstap_latency0_1_total Number of queries answered in less than 1ms" )
        p.append( "# TYPE dnstap_latency0_1_total counter" )
        p.append( "dnstap_latency0_1 %s" % counters["response/latency0_1"] )
        
        p.append( "# HELP dnstap_latency1_10_total Number of queries answered in 1-10 ms" )
        p.append( "# TYPE dnstap_latency1_10_total counter" )
        p.append( "dnstap_latency1_10 %s" % counters["response/latency1_10"] )
        
        p.append( "# HELP dnstap_latency10_50_total Number of queries answered in 10-50 ms" )
        p.append( "# TYPE dnstap_latency10_50_total counter" )
        p.append( "dnstap_latency10_50 %s" % counters["response/latency10_50"] )
        
        p.append( "# HELP dnstap_latency50_100_total Number of queries answered in 50-100 ms" )
        p.append( "# TYPE dnstap_latency50_100_total counter" )
        p.append( "dnstap_latency50_100 %s" % counters["response/latency50_100"] )
        
        p.append( "# HELP dnstap_latency100_1000_total Number of queries answered in 100-1000 ms" )
        p.append( "# TYPE dnstap_latency100_1000_total counter" )
        p.append( "dnstap_latency100_1000 %s" % counters["response/latency100_1000"] )
        
        p.append( "# HELP dnstap_latency_slow_total Number of queries answered in more than 1 second" )
        p.append( "# TYPE dnstap_latency_slow_total counter" )
        p.append( "dnstap_latency_slow %s" % counters["response/latency_slow"] )
        
        for s in self.stats.get_nameslist():
            sub_cntrs = self.stats.get_counters(stream=s, filters=filters)
            
            p.append( "# HELP dnstap_queries_bytes Total number of bytes for queries of this dnstap identity" )
            p.append( "# TYPE dnstap_queries_bytes counter" )
        
            p.append( "# HELP dnstap_responses_bytes Total number of bytes for responses of this dnstap identity" )
            p.append( "# TYPE dnstap_responses_bytes counter" )
        
            p.append( "# HELP dnstap_pps Number of packets per second for this dnstap identity" )
            p.append( "# TYPE dnstap_pps gauge" )
            
            p.append( "# HELP dnstap_qps Number of queries per second for this dnstap identity" )
            p.append( "# TYPE dnstap_qps gauge" )
            
            p.append( "# HELP dnstap_rps Number of responses per second for this dnstap identity" )
            p.append( "# TYPE dnstap_rps gauge" )
            
            p.append( "# HELP dnstap_domains Number of domains asked for this dnstap identity" )
            p.append( "# TYPE dnstap_domains counter" )
        
            p.append( "# HELP dnstap_clients Number of clients asked for this dnstap identity" )
            p.append( "# TYPE dnstap_clients counter" )

            p.append( "# HELP dnstap_queries Number of queries received for this dnstap identity" )
            p.append( "# TYPE dnstap_queries counter" )
            
            p.append( "# HELP dnstap_queries_udp Number of UDP queries received for this dnstap identity" )
            p.append( "# TYPE dnstap_queries_udp counter" )
            
            p.append( "# HELP dnstap_queries_tcp Number of TCP queries received for this dnstap identity" )
            p.append( "# TYPE dnstap_queries_tcp counter" )
            
            p.append( "# HELP dnstap_queries_inet Number of IPv4 queries received for this dnstap identity" )
            p.append( "# TYPE dnstap_queries_inet counter" )
            
            p.append( "# HELP dnstap_queries_inet6 Number of IPv6 queries received for this dnstap identity" )
            p.append( "# TYPE dnstap_queries_inet6 counter" )

            p.append( "# HELP dnstap_queries_a Number of A queries received for this dnstap identity" )
            p.append( "# TYPE dnstap_queries_a counter" )
            
            p.append( "# HELP dnstap_queries_aaaa Number of AAAA queries received for this dnstap identity" )
            p.append( "# TYPE dnstap_queries_aaaa counter" )
            
            p.append( "# HELP dnstap_queries_svr Number of SVR queries received for this dnstap identity" )
            p.append( "# TYPE dnstap_queries_svr counter" )

            p.append( "# HELP dnstap_responses Number of responses received for this dnstap identity" )
            p.append( "# TYPE dnstap_responses counter" )
            
            p.append( "# HELP dnstap_response_udp Number of UDP responses received for this dnstap identity" )
            p.append( "# TYPE dnstap_response_udp counter" )
            
            p.append( "# HELP dnstap_responses_tcp Number of TCP responses received for this dnstap identity" )
            p.append( "# TYPE dnstap_response_tcp counter" )
            
            p.append( "# HELP dnstap_response_inet Number of IPv4 responses received for this dnstap identity" )
            p.append( "# TYPE dnstap_response_inet counter" )
            
            p.append( "# HELP dnstap_response_inet6 Number of IPv6 responses received for this dnstap identity" )
            p.append( "# TYPE dnstap_response_inet6 counter" )
            
            p.append( "# HELP dnstap_responses_noerror Number of NOERROR answers for this dnstap identity" )
            p.append( "# TYPE dnstap_responses_noerror counter" )
            
            p.append( "# HELP dnstap_responses_nxdomain Number of NXDomain answers for this dnstap identity" )
            p.append( "# TYPE dnstap_responses_nxdomain counter" )
            
            p.append( "# HELP dnstap_responses_serverfail Number of SERVFAIL answers for this dnstap identity" )
            p.append( "# TYPE dnstap_responses_serverfail counter" )
            
            p.append( "# HELP dnstap_responses_refused Number of REFUSED answers for this dnstap identity" )
            p.append( "# TYPE dnstap_responses_refused counter" )
        
            p.append( "# HELP dnstap_latency1_10 Number of queries answered in 1-10 ms for this dnstap identity" )
            p.append( "# TYPE dnstap_latency1_10 counter" )
            
            p.append( "# HELP dnstap_latency1_10 Number of queries answered in 1-10 ms for this dnstap identity" )
            p.append( "# TYPE dnstap_latency1_10 counter" )
        
            p.append( "# HELP dnstap_latency10_50 Number of queries answered in 10-50 ms for this dnstap identity" )
            p.append( "# TYPE dnstap_latency10_50 counter" )
        
            p.append( "# HELP dnstap_latency10_50 Number of queries answered in 10-50 ms for this dnstap identity" )
            p.append( "# TYPE dnstap_latency10_50 counter" )
        
            p.append( "# HELP dnstap_latency100_1000 Number of queries answered in 100-1000 ms for this dnstap identity" )
            p.append( "# TYPE dnstap_latency100_1000 counter" )
            
            p.append( "# HELP dnstap_latency_slow Number of queries answered in more than 1 second for this dnstap identity" )
            p.append( "# TYPE dnstap_latency_slow counter" )


            p.append( "dnstap_queries_bytes{identity=\"%s\"} %s" % (s,sub_cntrs["query/bytes"]) )
            p.append( "dnstap_responses_bytes{identity=\"%s\"} %s" % (s,sub_cntrs["response/bytes"]) )
            
            p.append( "dnstap_pps{identity=\"%s\"} %s" % (s,sub_cntrs["pps"]) )
            p.append( "dnstap_qps{identity=\"%s\"} %s" % (s,sub_cntrs["qps"]) )
            p.append( "dnstap_rps{identity=\"%s\"} %s" % (s,sub_cntrs["rps"]) )
            p.append( "dnstap_domains{identity=\"%s\"} %s" % (s,sub_cntrs["domains"]) )
            p.append( "dnstap_clients{identity=\"%s\"} %s" % (s,sub_cntrs["clients"]) )
            
            p.append( "dnstap_queries{identity=\"%s\"} %s" % (s,sub_cntrs["query"]) )
            p.append( "dnstap_queries_udp{identity=\"%s\"} %s" % (s,sub_cntrs["query/udp"]) )
            p.append( "dnstap_queries_tcp{identity=\"%s\"} %s" % (s,sub_cntrs["query/tcp"]) )
            p.append( "dnstap_queries_inet{identity=\"%s\"} %s" % (s,sub_cntrs["query/inet"]) )
            p.append( "dnstap_queries_inet6{identity=\"%s\"} %s" % (s,sub_cntrs["query/inet6"]) )
            p.append( "dnstap_queries_a{identity=\"%s\"} %s" % (s,sub_cntrs["query/a"]) )
            p.append( "dnstap_queries_aaaa{identity=\"%s\"} %s" % (s,sub_cntrs["query/aaaa"]) )
            p.append( "dnstap_queries_svr{identity=\"%s\"} %s" % (s,sub_cntrs["query/svr"]) )

            p.append( "dnstap_responses{identity=\"%s\"} %s" % (s,sub_cntrs["response"]) )
            p.append( "dnstap_response_udp{identity=\"%s\"} %s" % (s,sub_cntrs["response/udp"]) )
            p.append( "dnstap_response_tcp{identity=\"%s\"} %s" % (s,sub_cntrs["response/tcp"]) )
            p.append( "dnstap_response_inet{identity=\"%s\"} %s" % (s,sub_cntrs["response/inet"]) )
            p.append( "dnstap_response_inet6{identity=\"%s\"} %s" % (s,sub_cntrs["response/inet6"]) )
            p.append( "dnstap_responses_noerror{identity=\"%s\"} %s" % (s,sub_cntrs["response/noerror"]) )
            p.append( "dnstap_responses_nxdomain{identity=\"%s\"} %s" % (s,sub_cntrs["response/nxdomain"]) )
            p.append( "dnstap_responses_servfail{identity=\"%s\"} %s" % (s,sub_cntrs["response/servfail"]) )
            p.append( "dnstap_responses_refused{identity=\"%s\"} %s" % (s,sub_cntrs["response/refused"]) )
            
            p.append( "dnstap_latency0_1{identity=\"%s\"} %s" % (s,sub_cntrs["response/latency0_1"]) )
            p.append( "dnstap_latency1_10{identity=\"%s\"} %s" % (s,sub_cntrs["response/latency1_10"]) )
            p.append( "dnstap_latency10_50{identity=\"%s\"} %s" % (s,sub_cntrs["response/latency10_50"]) )
            p.append( "dnstap_latency50_100{identity=\"%s\"} %s" % (s,sub_cntrs["response/latency50_100"]) )
            p.append( "dnstap_latency100_1000{identity=\"%s\"} %s" % (s,sub_cntrs["response/latency100_1000"]) )
            p.append( "dnstap_latency_slow{identity=\"%s\"} %s" % (s,sub_cntrs["response/latency_slow"]) )
            
        return web.Response(text="\n".join(p), content_type='text/plain')
        
    async def handle_counters(self, request):
        auth = self.check_auth(request=request)
        if not auth:
            return web.Response(status=401)

        #n = request.query.get("n", self.top)
        s = request.query.get("stream", None)
        more_filters = request.query.get("more", [])
        if not isinstance(more_filters, list):
            more_filters = more_filters.split(",")
            
        filters = self.cfg_stats["default-counters"]
        filters.extend(more_filters)
        
        data = {"stream": s, "counters": self.stats.get_counters(s, filters=filters)}
        return web.json_response(data)
        
    async def handle_tables(self, request):
        auth = self.check_auth(request=request)
        if not auth:
            return web.Response(status=401)

        n = request.query.get("n", self.top)
        s = request.query.get("stream", None)
        more_filters = request.query.get("more", [])
        if not isinstance(more_filters, list):
            more_filters = more_filters.split(",")
            
        filters = self.cfg_stats["default-top"]
        filters.extend(more_filters)

        data = { "stream": s,
                 "tables": {
                     "tlds": self.stats.top_tlds(int(n),s),
                     "domains": self.stats.top_domains(int(n),s, filters=filters),
                     "clients": self.stats.top_clients(int(n), s),
                     "rcodes": self.stats.top_rcodes(int(n), s),
                     "rrtypes": self.stats.top_rrtypes(int(n),s) }
                }
        return web.json_response(data)
        
    async def handle_streams(self, request):
        auth = self.check_auth(request=request)
        if not auth:
            return web.Response(status=401)

        data = { "streams": self.stats.get_nameslist() }
        return web.json_response(data)
        
async def create_server(loop, cfg, stats, cfg_stats):
    # api ressources
    basic_auth = BasicAuth(login=cfg["login"], password=cfg["password"])
    hdlrs = Handlers(cfg["api-key"], basic_auth, stats, cfg_stats)
    
    # rest api server
    app = web.Application(loop=loop)
 
    # endpoints
    app.router.add_get('/metrics', hdlrs.handle_metrics)
    app.router.add_get('/tables', hdlrs.handle_tables)
    app.router.add_get('/counters', hdlrs.handle_counters)
    app.router.add_get('/streams', hdlrs.handle_streams)
    app.router.add_delete('/reset', hdlrs.handle_reset)

    # create the server
    listen_address = (cfg["local-address"], cfg["local-port"])
    try:
        srv = await loop.create_server(app.make_handler(access_log=None), *listen_address)
    except OSError as e:
        clogger.error( "http api: %s" % e)
        return None
        
    clogger.debug("Webserver: listening on %s:%s" % listen_address )
    
    return srv




import re
import logging
import socket
import hashlib
import struct
import dnstap_pb

import dnstap_receiver.dns.rdatatype as dns_rdatatypes
import dnstap_receiver.dns.rcode as dns_rcodes
import dnstap_receiver.dns.parser as dns_parser

# create default logger for the dnstap receiver
clogger = logging.getLogger("dnstap_receiver.console")

DNSTAP_TYPE = dnstap_pb.dnstap_pb2._MESSAGE_TYPE.values_by_number
DNSTAP_FAMILY = dnstap_pb.dnstap_pb2._SOCKETFAMILY.values_by_number
DNSTAP_PROTO = dnstap_pb.dnstap_pb2._SOCKETPROTOCOL.values_by_number  

class UnknownValue:
    name = "-"

async def cb_ondnstap(dnstap_decoder, payload, cfg, queues_list, stats, geoip_reader, cache):
    """on dnstap"""
    # decode binary payload
    dnstap_decoder.ParseFromString(payload)
    dm = dnstap_decoder.message
            
    # type: CLIENT_QUERY
    # socket_family: INET
    # socket_protocol: UDP
    # query_address: "\n\000\000\002"
    # response_address: "\n\000\000\322"
    # query_port: 33019
    # response_port: 53
    # query_time_sec: 1609271575
    # query_time_nsec: 779179701
    # query_message: "\2300\001 \000\001\000\000\000\000\000\001\003www\006google\003com\000\000\034\000
    # \001\000\000)\020\000\000\000\000\000\000\014\000\n\000\010U\222\\\270\340\330jg"

    # type: CLIENT_RESPONSE
    # socket_family: INET
    # socket_protocol: UDP
    # query_address: "\n\000\000\002"
    # response_address: "\n\000\000\322"
    # query_port: 33019
    # response_port: 53
    # query_time_sec: 1609271575
    # query_time_nsec: 779179701
    # response_time_sec: 1609271575
    # response_time_nsec: 831279572
    # response_message: "\2300\201\200\000\001\000\001\000\000\000\001\003www\006google\003com\000\000\034
    # \000\001\300\014\000\034\000\001\000\000\000\236\000\020*\000\024P@\007\010\006\000\000\000\000\000\000 \004\000\000)\004\320\000\000\000\000\000\000"
    
    tap = { "identity": UnknownValue.name, 
            "qname": UnknownValue.name, 
            "rrtype": UnknownValue.name, 
            "query-ip": UnknownValue.name, "query-port": UnknownValue.name,
            "response-ip": UnknownValue.name, "response-port": UnknownValue.name,
            "latency": UnknownValue.name}

    # filtering by dnstap identity ?
    if len(dnstap_decoder.identity): tap["identity"] = dnstap_decoder.identity.decode()
    if cfg["filter"]["dnstap-identities"] is not None:
        if re.match(cfg["filter"]["dnstap-identities"], tap["identity"]) is None:
            return

    # decode type message
    tap["message"] = DNSTAP_TYPE.get(dm.type, UnknownValue.name).name
    tap["family"] = DNSTAP_FAMILY.get(dm.socket_family, UnknownValue.name).name
    tap["protocol"] = DNSTAP_PROTO.get(dm.socket_protocol, UnknownValue.name).name

    # decode query address
    qaddr = dm.query_address
    if len(qaddr) and dm.socket_family == 1:
        qaddr = qaddr[12:] if len(qaddr) == 16 else qaddr # condition for coredns, address is 16 bytes long so keept only 4 bytes
        tap["query-ip"] = socket.inet_ntop(socket.AF_INET, qaddr) # socket.inet_ntoa(qaddr)
    if len(qaddr) and dm.socket_family == 2: tap["query-ip"] = socket.inet_ntop(socket.AF_INET6, qaddr)
    if dm.query_port > 0: tap["query-port"] = dm.query_port

    # decode response address
    raddr = dm.response_address
    if len(raddr) and dm.socket_family == 1:
        raddr = raddr[12:] if len(raddr) == 16 else raddr # condition for coredns, address is 16 bytes long so keept only 4 bytes
        tap["response-ip"] = socket.inet_ntop(socket.AF_INET, raddr) # socket.inet_ntoa(qaddr)
    if len(raddr) and dm.socket_family == 2: tap["response-ip"] = socket.inet_ntop(socket.AF_INET6, raddr)
    if dm.response_port > 0: tap["response-port"] = dm.response_port
    
    # decode dns message
    dns_payload = dm.query_message if (dm.type % 2 ) == 1 else dm.response_message
    tap["payload"] = dns_payload
    
    if len(dns_payload) >= dns_parser.DNS_LEN:
        dns_id, dns_rcode, dns_qdcount = dns_parser.decode_dns(dns_payload)
    else:
        clogger.error('Dnstap decoder - dns payload too short: %s' % dns_payload)
        return
        
    if (dm.type % 2 ) == 1 :               
        tap["length"] = len(dm.query_message)
        tap["timestamp"] = dm.query_time_sec + round(dm.query_time_nsec )*1e-9
        tap["time-sec"] = dm.query_time_sec
        tap["time-nsec"] = dm.query_time_nsec
        tap["type"] = "query"
        
        # hash query and put in cache the arrival time
        if len(dm.query_address) and dm.query_port > 0:
            hash_payload = "%s+%s+%s" % (dm.query_address, str(dm.query_port), dns_id)
            qhash = hashlib.sha1(hash_payload.encode()).hexdigest()
            cache[qhash] = tap["timestamp"]
            
    # handle response message
    if (dm.type % 2 ) == 0 :
        tap["length"] = len(dm.response_message)
        tap["timestamp"] = dm.response_time_sec + round(dm.response_time_nsec )*1e-9
        tap["time-sec"] = dm.response_time_sec
        tap["time-nsec"] = dm.response_time_nsec        
        tap["type"] = "response"

        # compute hash of the query and latency
        if len(dm.query_address) and dm.query_port > 0:
            hash_payload = "%s+%s+%s" % (dm.query_address, str(dm.query_port), dns_id)
            qhash = hashlib.sha1(hash_payload.encode()).hexdigest()
            if qhash in cache: tap["latency"] = round(tap["timestamp"]-cache[qhash],3)

    # common params
    if dns_qdcount:
        qname, qtype = dns_parser.decode_question(dns_payload)
        tap["qname"] = qname.decode(errors="ignore")
        tap["rrtype"] = dns_rdatatypes.RDATATYPES[qtype]
        
    tap["rcode"] = dns_rcodes.RCODES[dns_rcode]
    tap["id"] = dns_id

    # filtering by qname ?
    if cfg["filter"]["qname-regex"] is not None:
        if re.match(cfg["filter"]["qname-regex"], tap["qname"]) is None:
            return

    # geoip support 
    if geoip_reader is not None:
        try:
            response = geoip_reader.city(tap["query-ip"])
            if cfg["geoip"]["country-iso"]:
                tap["country"] = response.country.iso_code
            else:
                tap["country"] = response.country.name
            if response.city.name is not None:
                tap["city"] = response.city.name
            else:
                tap["city"] = UnknownValue.name
        except Exception as e:
            tap["country"] = UnknownValue.name
            tap["city"] = UnknownValue.name
            
    # update metrics 
    stats.record(tap=tap)
        
    # append the dnstap message to the queue
    for q in queues_list:
        q.put_nowait(tap.copy())



import socket
import struct
import logging
import asyncio
import time
import hashlib

import dnstap_receiver.dns.rdatatype as dns_rdatatypes
import dnstap_receiver.dns.rcode as dns_rcodes

clogger = logging.getLogger("dnstap_receiver.console")

SO_TIMESTAMPNS = 35

ETH_LEN     = 14
IPV4_LEN    = 20
IPV6_LEN    = 40 # 40 bytes
UDP_LEN     = 8
DNS_LEN     = 12
TCP_LEN     = 20 # header length in bytes

IPv4        = 0x0800
IPv6        = 0x86DD

TCP         = 6
UDP         = 17

def get_socket(interface):
    s = socket.socket(socket.AF_PACKET, socket.SOCK_RAW, socket.SOCK_RAW)
    s.setsockopt(socket.SOL_SOCKET, SO_TIMESTAMPNS, 1)
    
    s.bind( (interface, socket.SOCK_RAW) )
    s.setblocking(False)
    clogger.debug("Input handler: listening on interface %s" % interface)
    
    return s

def decode_question(dns_tap, data):
    buf = data
    qname = []
    
    while len(buf):
        length = buf[0]
        if length == 0x00:
            break
        label = buf[1:length+1]
        qname.append(buf[1:length+1])
        buf = buf[length+1:]

    dns_tap["rrtype"] = dns_rdatatypes.RDATATYPES[int.from_bytes(buf[1:3], "big")]
    dns_tap["qname"] = b".".join(qname) + b"."
    dns_tap["qname"] = dns_tap["qname"].decode()
    
async def watch_buffer(cfg, q, queues_list, stats, cache, start_shutdown):
    loop = asyncio.get_event_loop()
    
    while not start_shutdown.is_set():
        p = await q.get()
        data, ancdata, addr = p
        
        _, proto, _, _, _ = addr # interface name, protocol, packet type, arp, phy
        _, _, cmsg_data = ancdata # cmsg_level, cmsg_type, cmsg_data
        
        tap = {}
        tap["identity"] = cfg["dnstap-identity"]
        tap["latency"] = "-"
        tap["qname"] = "-"
        tap["rrtype"] = "-"
        tap["query-ip"] = "-"
        tap["query-port"] = "-"
        
        # extract arrival time
        tsec, _, nsec, _ = struct.unpack("iiii",cmsg_data)
        tap["time-sec"] = tsec
        tap["time-nsec"] = nsec
        tap["timestamp"] = tsec + nsec*1e-10
        
        # extract ethernet data 
        eth_data = data[ETH_LEN:]

        # decode ipv4 header ?
        if proto == IPv4:
            ip = eth_data[:IPV4_LEN]
            
            ip_proto = ip[9]
            ip_ihl = ( ip[0] & 15 )
            ip_data = eth_data[ (ip_ihl*32)//8: ]
            
            tap["family"] = "INET"
            tap["src-ip"] = socket.inet_ntoa(ip[12:16])
            tap["dst-ip"] = socket.inet_ntoa(ip[16:])
        
        # decode ipv6 header ?
        elif proto == IPv6:
            ip = eth_data[:IPV6_LEN]

            ip_proto = ip[6]
            ip_data = eth_data[IPV6_LEN:]
            
            tap["family"] = "INET6"
            tap["src-ip"] = socket.inet_ntop(socket.AF_INET6, ip[8:24])
            tap["dst-ip"] = socket.inet_ntop(socket.AF_INET6, ip[24:])

        else:
            continue # ignore other packet than ip

        # decode udp header ?
        if ip_proto == UDP:
            tap["protocol"] = "UDP"
            udp = ip_data[:UDP_LEN]

            tap["src-port"] = int.from_bytes(udp[0:2], "big")
            tap["dst-port"] = int.from_bytes(udp[2:4], "big")

            dns_payload = ip_data[UDP_LEN:]
        
        # decode tcp header ?
        # fragmented dns payload in several tcp push not yet implemented!
        elif ip_proto == TCP:
            tap["protocol"] = "TCP"
            
            tcp = ip_data[:TCP_LEN]

            psh = (int.from_bytes(tcp[12:14], "big") & 8) and 1 or 0
            if not psh: continue
            
            tap["src-port"] = int.from_bytes(tcp[0:2], "big")
            tap["dst-port"] = int.from_bytes(tcp[2:4], "big")
            
            data_offset = ( int.from_bytes(tcp[12:14], "big")  >> 12)*32//8
            dns_tcp_payload = ip_data[data_offset:]
        else:
            continue # ignore other protocol than udp and tcp

        # ignore packets according to the port
        if tap["src-port"] not in cfg["dns-port"] and tap["dst-port"] not in cfg["dns-port"]:
            continue

        if ip_proto == TCP:
            dns_length = int.from_bytes(dns_tcp_payload[0:2], "big")
            dns_payload = dns_tcp_payload[2:]
            
        # ignore the packet if the dns payload if too small
        if len(dns_payload) < DNS_LEN:  continue
 
        tap["payload"] = dns_payload
        # begin to decode the dns payload
        tap["id"] = int.from_bytes(dns_payload[0:2], "big")
        tap["type"] = "response" if int.from_bytes(dns_payload[2:4], "big") >> 15 else "query"
        
        tap["opcode"] = (int.from_bytes(dns_payload[2:4], "big") & 0x7800) >> 11
        tap["aa"] = (int.from_bytes(dns_payload[2:4], "big") & 0x0400) != 0 # authoritative answer
        tap["rd"] = (int.from_bytes(dns_payload[2:4], "big") & 0x100) != 0 # recursion desired
        tap["ra"] = (int.from_bytes(dns_payload[2:4], "big") & 0x80) != 0 # recursion available
        tap["tc"] = (int.from_bytes(dns_payload[2:4], "big") & 0x200) != 0 # truncated answer

        # find the type of dnstap message ?
        if cfg["client-query-support"] and tap["type"] == "query" and (tap["dst-ip"] in cfg["eth-ip"]):
            tap["message"] = "CLIENT_QUERY"
            tap["query-ip"] = tap["src-ip"]
            tap["query-port"] = tap["src-port"]
            tap["response-ip"] = tap["dst-ip"]
            tap["response-port"] = tap["dst-port"]
            
        elif cfg["client-response-support"] and tap["type"] == "response" and (tap["src-ip"] in cfg["eth-ip"]):
            tap["message"] = "CLIENT_RESPONSE"
            tap["query-ip"] = tap["dst-ip"]
            tap["query-port"] = tap["dst-port"]
            tap["response-ip"] = tap["src-ip"]
            tap["response-port"] = tap["src-port"]
            
        elif cfg["resolver-query-support"] and tap["type"] == "query" and not tap["rd"] and (tap["src-ip"] in cfg["eth-ip"]):
            tap["message"] = "RESOLVER_QUERY"
            tap["query-ip"] = tap["src-ip"]
            tap["query-port"] = tap["src-port"]
            tap["response-ip"] = tap["dst-ip"]
            tap["response-port"] = tap["dst-port"]
            
        elif cfg["resolver-response-support"] and tap["type"] == "response" and tap["aa"] and (tap["dst-ip"] in cfg["eth-ip"]):
            tap["message"] = "RESOLVER_RESPONSE"
            tap["query-ip"] = tap["dst-ip"]
            tap["query-port"] = tap["dst-port"]
            tap["response-ip"] = tap["src-ip"]
            tap["response-port"] = tap["src-port"]

        elif cfg["forwarder-query-support"] and tap["type"] == "query" and tap["rd"] and (tap["src-ip"] in cfg["eth-ip"]):
            tap["message"] = "FORWARDER_QUERY"
            tap["query-ip"] = tap["src-ip"]
            tap["query-port"] = tap["src-port"]
            tap["response-ip"] = tap["dst-ip"]
            tap["response-port"] = tap["dst-port"]
            
        elif cfg["forwarder-response-support"] and tap["type"] == "response" and tap["rd"] and (tap["dst-ip"] in cfg["eth-ip"]):
            tap["message"] = "FORWARDER_RESPONSE"
            tap["query-ip"] = tap["dst-ip"]
            tap["query-port"] = tap["dst-port"]
            tap["response-ip"] = tap["src-ip"]
            tap["response-port"] = tap["src-port"]
            
        else:
            continue

        # compute latency
        if tap["type"] == "query":
            hash_payload = "%s+%s+%s" % (tap["query-ip"], tap["src-port"], tap["id"])
            qhash = hashlib.sha1(hash_payload.encode()).hexdigest()
            cache[qhash] = tap["timestamp"]

        if tap["type"] == "response":
            hash_payload = "%s+%s+%s" % (tap["query-ip"], str(tap["dst-port"]), tap["id"])
            qhash = hashlib.sha1(hash_payload.encode()).hexdigest()
            if qhash in cache: tap["latency"] = round(tap["timestamp"]-cache[qhash],3)
          
        tap["length"] = len(dns_payload)
        
        tap["rcode"] = dns_rcodes.RCODES[int.from_bytes(dns_payload[2:4], "big") & 15]
        dns_qdcount = int.from_bytes(dns_payload[4:6], "big")
        
        # decode the question ?
        if dns_qdcount: decode_question(tap, dns_payload[DNS_LEN:])

        # update metrics 
        stats.record(tap=tap)
            
        # append the dnstap message to the queue
        for q_out in queues_list:
            q_out.put_nowait(tap)
            
def start_input(cfg, queue_sniffer, start_shutdown):
    clogger.debug("Input handler: sniffer")

    # prepare the socket
    s = get_socket(interface=cfg["eth-name"])

    # read data
    bufsize = 65535
    
    while not start_shutdown.is_set():
        try:
            raw_data, ancdata, _, address = s.recvmsg(bufsize, 1024)
        except BlockingIOError:
            time.sleep(0.01)
            continue
        queue_sniffer.put_nowait( (raw_data, ancdata[0], address) )
    
    clogger.debug("Input handler: closing sniffer")
    s.close()



import asyncio
import logging

clogger = logging.getLogger("dnstap_receiver.console")

from dnstap_receiver.outputs import transform

def checking_conf(cfg):
    """validate the config"""
    clogger.debug("Output handler: syslog")
    
    valid_conf = True
    
    if cfg["remote-address"] is None:
        valid_conf = False
        clogger.error("Output handler: no remote address provided")
        
    if cfg["remote-port"] is None:
        valid_conf = False
        clogger.error("Output handler: no port provided")
            
    return valid_conf
    
async def syslog_tcpclient(output_cfg, queue, start_shutdown):
    host, port = output_cfg["remote-address"], output_cfg["remote-port"]
    clogger.debug("Output handler: connection to %s:%s" % (host,port) )
    reader, tcp_writer = await asyncio.open_connection(host, port)
    clogger.debug("Output handler: connected")
    
    # consume queue
    while not start_shutdown.is_set():
        # read item from queue
        tapmsg = await queue.get()
        
        # convert dnstap message
        msg = transform.convert_dnstap(fmt=output_cfg["format"], tapmsg=tapmsg)
            
        # add count octets
        s_msg = "%s" % len(msg)
        tcp_writer.write( b"%s %s" % (s_msg.encode(), msg) )
        
        # connection lost ? exit and try to reconnect 
        if tcp_writer.transport._conn_lost:
            break
        
        # done continue to next item
        queue.task_done()
        
    # something 
    if not start_shutdown.is_set():
        clogger.error("Output handler: connection lost")
 
async def handle(output_cfg, queue, metrics, start_shutdown):
    """handle output"""
    server_address = (output_cfg["remote-address"], output_cfg["remote-port"])
    loop = asyncio.get_event_loop()
    
    # syslog tcp
    if output_cfg["transport"] == "tcp":
        clogger.debug("Output handler: syslog TCP enabled")
        while not start_shutdown.is_set():
            try:
                await syslog_tcpclient(output_cfg, queue, start_shutdown)
            except ConnectionRefusedError:
                clogger.error('Output handler: connection to syslog server failed!')
            except asyncio.TimeoutError:
                clogger.error('Output handler: connection to syslog server timed out!')
            else:
                clogger.error('Output handler: connection to server is closed.')
                
            clogger.debug("'Output handler: retry to connect every %ss" % output_cfg["retry"])
            if not start_shutdown.is_set():
                await asyncio.sleep(output_cfg["retry"])
    
    # syslog udp
    else:
        clogger.debug("Output handler: syslog UDP enabled with %s" % str(server_address) )
        transport, _  = await loop.create_datagram_endpoint(asyncio.DatagramProtocol,
                                                            remote_addr=server_address)

        # consume the queue
        while not start_shutdown.is_set():
            # read item from queue
            tapmsg = await queue.get()
            
            # convert dnstap message
            msg = transform.convert_dnstap(fmt=output_cfg["format"], tapmsg=tapmsg)
        
            # send msg
            transport.sendto( msg )
            
            # all done
            queue.task_done()



import logging
import asyncio
import sys

clogger = logging.getLogger("dnstap_receiver.console")
metrics_logger = logging.getLogger("dnstap_receiver.output.metrics")

def checking_conf(cfg):
    """validate the config"""
    clogger.debug("Output handler: metrics")
    return True
    
def setup_logger(cfg):
    """setup loggers"""
    logfmt = '%(asctime)s %(message)s'
    max_bytes = int(cfg["file-max-size"].split('M')[0]) * 1024 * 1024
    
    metrics_logger.setLevel(logging.INFO)
    metrics_logger.propagate = False
    
    if cfg["file"] is not None:
        lh = logging.handlers.RotatingFileHandler(
            cfg["file"],
            maxBytes=max_bytes,
            backupCount=cfg["file-count"]
        )
    else:
        lh = logging.StreamHandler(stream=sys.stdout)
    lh.setLevel(logging.INFO)
    lh.setFormatter(logging.Formatter(logfmt))    
    
    metrics_logger.addHandler(lh)
    
    
async def handle(cfg, queue, metrics, start_shutdown):
    """stdout output handler"""
    # init logger
    setup_logger(cfg)

    shutdown_wait_task = asyncio.create_task(start_shutdown.wait())
    sleep_task = asyncio.create_task(asyncio.sleep(cfg["interval"]))
    while True:
        done, pending = await asyncio.wait(
            [shutdown_wait_task, sleep_task],
            return_when=asyncio.FIRST_COMPLETED,
        )

        if shutdown_wait_task in done:
            sleep_task.cancel()
            return
        else:
            sleep_task = asyncio.sleep(cfg["interval"])

        # clear queue
        for _ in range(queue.qsize()):
            queue.get_nowait()
            queue.task_done()
            
        # get counters
        filters = [ "query", "qps", "clients", "domains",
                    "query/inet", "query/inet6", 
                    "query/udp", "query/tcp",
                    "response/noerror", "response/nxdomain",
                    "query/a", "query/aaaa"] 
        counters = metrics.get_counters(filters=filters)

        msg = [ "%s QUERIES" % counters["query"] ]
        msg.append( "%s QPS" % counters["qps"] )
        
        msg.append( "%s DOMAINS" % counters["domains"] )
        msg.append( "%s CLIENTS" % counters["clients"] )
        
        msg.append( "%s INET" % counters["query/inet"] )   
        msg.append( "%s INET6" % counters["query/inet6"] )

        msg.append( "%s UDP" % counters["query/udp"] )
        msg.append( "%s TCP" % counters["query/tcp"] )
        
        msg.append( "%s NOERROR" % counters["response/noerror"] )
        msg.append( "%s NXDOMAINS" % counters["response/nxdomain"] )
        
        msg.append( "%s A" % counters["query/a"] )
        msg.append( "%s AAAA" % counters["query/aaaa"] )
        
        # print to stdout
        metrics_logger.info(", ".join(msg))
        
        # reset stats?
        if not cfg["cumulative"]:
            metrics.reset()



import asyncio
import logging
import socket
import fstrm
import dnstap_pb

from dnstap_receiver.outputs import transform

clogger = logging.getLogger("dnstap_receiver.console")

def checking_conf(cfg):
    """validate the config"""
    clogger.debug("Output handler: dnstap")
    
    valid_conf = True
    
    if cfg["remote-address"] is None:
        valid_conf = False
        clogger.error("Output handler: no remote address provided")
        
    if cfg["remote-port"] is None:
        valid_conf = False
        clogger.error("Output handler: no port provided")
            
    return valid_conf
    
async def dnstap_client(output_cfg, queue, start_shutdown):
    host, port = output_cfg["remote-address"], output_cfg["remote-port"]
    clogger.debug("Output handler: connection to %s:%s" % (host,port) )
    reader, tcp_writer = await asyncio.open_connection(host, port)
    clogger.debug("Output handler: connected")
    
    content_type = b"protobuf:dnstap.Dnstap"
    fstrm_handler = fstrm.FstrmCodec()
    dnstap = dnstap_pb.Dnstap()
    
    # framestream - do handshake 
    ctrl_ready = fstrm_handler.encode(ctrl=fstrm.FSTRM_CONTROL_READY, ct=[content_type])
    tcp_writer.write(ctrl_ready)
    
    while not start_shutdown.is_set():
        data = await reader.read(fstrm_handler.pending_nb_bytes())
        if not len(data):
            break
        fstrm_handler.append(data=data)
        
        # process the buffer, check if we have received a complete frame ?
        if fstrm_handler.process():
            # Ok, the frame is complete so let's decode it
            ctrl, ct, payload  = fstrm_handler.decode()
                
            if ctrl == fstrm.FSTRM_CONTROL_ACCEPT:
                ctrl_start = fstrm_handler.encode(ctrl=fstrm.FSTRM_CONTROL_START)
                tcp_writer.write(ctrl_start)
                break

    if start_shutdown.is_set():
        return

    # consume queue and send data frames
    while not start_shutdown.is_set():
        # read item from queue
        tap = await queue.get()
        
        dnstap.Clear()
        
        dnstap.type = 1
        dnstap.version = b"-"
        dnstap.identity = output_cfg["dnstap-identity"].encode()
        
        dnstap.message.type = dnstap_pb.dnstap_pb2._MESSAGE_TYPE.values_by_name[tap["message"]].number
        dnstap.message.socket_protocol = dnstap_pb.dnstap_pb2._SOCKETPROTOCOL.values_by_name[tap["protocol"]].number
        dnstap.message.socket_family = dnstap_pb.dnstap_pb2._SOCKETFAMILY.values_by_name[tap["family"]].number

        if tap["type"] == "query":
            dnstap.message.query_message = tap["payload"]
            dnstap.message.query_time_nsec = tap["time-nsec"]
            dnstap.message.query_time_sec = tap["time-sec"]
        else:
            dnstap.message.response_message = tap["payload"]
            dnstap.message.response_time_nsec = tap["time-nsec"]
            dnstap.message.response_time_sec = tap["time-sec"]

        if tap["family"] == "INET":
            dnstap.message.query_address = socket.inet_pton(socket.AF_INET, tap["query-ip"])
            dnstap.message.response_address = socket.inet_pton(socket.AF_INET, tap["response-ip"])
            
        if tap["family"] == "INET6":
            dnstap.message.query_address = socket.inet_pton(socket.AF_INET6, tap["query-ip"])
            dnstap.message.response_address = socket.inet_pton(socket.AF_INET6, tap["response-ip"])
            
        dnstap.message.query_port = tap["query-port"]
        dnstap.message.response_port = tap["response-port"]

        # convert to dnstap message
        data = fstrm_handler.encode(ctrl=fstrm.FSTRM_DATA_FRAME, payload=dnstap.SerializeToString())
        tcp_writer.write(data)
        
        # connection lost ? exit and try to reconnect 
        if tcp_writer.transport._conn_lost:
            break
        
        # done continue to next item
        queue.task_done()

    if start_shutdown.is_set():
        return
        
    # something 
    clogger.error("Output handler: connection lost")
 
async def handle(output_cfg, queue, metrics, start_shutdown):
    """handle output"""
    server_address = (output_cfg["remote-address"], output_cfg["remote-port"])
    loop = asyncio.get_event_loop()
    
    clogger.debug("Output handler: DNS tap enabled")
    while not start_shutdown.is_set():
        try:
            await dnstap_client(output_cfg, queue, start_shutdown)
        except ConnectionRefusedError:
            clogger.error('Output handler: connection to remote dnstap receiver failed!')
        except TimeoutError:
            clogger.error('Output handler: connection to remote dnstap receiver timed out!')
        else:
            clogger.error('Output handler: connection to remote dnstap receiver is closed.')
            
        clogger.debug("Output handler: retry to connect every %ss" % output_cfg["retry"])
        if not start_shutdown.is_set():
            await asyncio.sleep(output_cfg["retry"])



import asyncio
import logging
import fstrm
import dnstap_pb

from dnstap_receiver.inputs import dnstap_decoder 

clogger = logging.getLogger("dnstap_receiver.console")

def checking_conf(cfg):
    """validate the config"""
    clogger.debug("Input handler: tcp client")
    
    valid_conf = True
    
    if cfg["remote-address"] is None:
        valid_conf = False
        clogger.error("Input handler: no remote address provided")
        
    if cfg["remote-port"] is None:
        valid_conf = False
        clogger.error("Input handler: no port provided")
            
    return valid_conf
    
async def tcp_client(cfg, cfg_input, queues_list, stats, geoip_reader, cache, start_shutdown):
    host, port = cfg_input["remote-address"], cfg_input["remote-port"]
    clogger.debug("Input handler: connection to %s:%s" % (host,port) )
    reader, tcp_writer = await asyncio.open_connection(host, port)
    clogger.debug("Input handler: connected")
    
    content_type = b"protobuf:dnstap.Dnstap"
    fstrm_handler = fstrm.FstrmCodec()
    dnstap = dnstap_pb.Dnstap()
 
                    
    try: 
        # framestream - do handshake 
        ctrl_ready = fstrm_handler.encode(ctrl=fstrm.FSTRM_CONTROL_READY, ct=[content_type])
        tcp_writer.write(ctrl_ready)
        
        shutdown_wait_task = asyncio.create_task(start_shutdown.wait())

        while True:
            read_task = asyncio.create_task(reader.read(fstrm_handler.pending_nb_bytes()))
            done, pending = await asyncio.wait(
                [shutdown_wait_task, read_task],
                return_when=asyncio.FIRST_COMPLETED,
            )

            if shutdown_wait_task in done:
                read_task.cancel()
                return
            else:
                shutdown_wait_task.cancel()
                data = await read_task

            if not len(data):
                break
            fstrm_handler.append(data=data)
            
            # process the buffer, check if we have received a complete frame ?
            if fstrm_handler.process():
                # Ok, the frame is complete so let's decode it
                ctrl, ct, payload  = fstrm_handler.decode()
                    
                if ctrl == fstrm.FSTRM_CONTROL_ACCEPT:
                    ctrl_start = fstrm_handler.encode(ctrl=fstrm.FSTRM_CONTROL_START)
                    tcp_writer.write(ctrl_start)
                    break

        # waiting for incoming data
        running = True
        while running:
            # read bytes
            read_task = asyncio.create_task(reader.read(fstrm_handler.pending_nb_bytes()))
            done, pending = await asyncio.wait(
                [shutdown_wait_task, read_task],
                return_when=asyncio.FIRST_COMPLETED,
            )

            if shutdown_wait_task in done:
                read_task.cancel()
                return
            else:
                data = await read_task

            if not len(data):
                running = False
                break
                
            # append data to the buffer
            fstrm_handler.append(data=data)
            
            # process the buffer, check if we have received a complete frame ?
            if fstrm_handler.process():
                # Ok, the frame is complete so let's decode it
                ctrl, ct, payload  = fstrm_handler.decode()

                # handle the DATA frame
                if ctrl == fstrm.FSTRM_DATA_FRAME:
                    loop.create_task(dnstap_decoder.cb_ondnstap(dnstap_protobuf, payload, cfg, queues_list, stats, geoip_reader, cache))
     
    except ConnectionError as e:
        clogger.error(f'Input handler: {peername} - %s' % e)
    except asyncio.CancelledError:
        clogger.debug(f'Input handler: {peername} - closing connection.')
        writer.close()
    except asyncio.IncompleteReadError:
        clogger.debug(f'Input handler: {peername} - disconnected')
    finally:
        writer.close()
        clogger.debug(f'Input handler: {peername} - closed')
        
async def start_tcpclient(cfg, queues_list, stats, geoip_reader, cache, start_shutdown):
    """start input tcp client"""
    server_address = (cfg_input["remote-address"], cfg_input["remote-port"])
    loop = asyncio.get_event_loop()
    
    clogger.debug("Input handler: TCP client enabled")
    while not start_shutdown.is_set():
        try:
            await tcp_client(cfg, cfg_input, queues_list, stats, geoip_reader, cache, start_shutdown)
        except ConnectionRefusedError:
            clogger.error('Input handler: connection to remote dns server failed!')
        except TimeoutError:
            clogger.error('Input handler: connection to remote dns server timed out!')
        else:
            clogger.error('Input handler: connection to remote dns server is closed.')
            
        clogger.debug("Input handler: retry to connect every %ss" % cfg_input["retry"])
        await asyncio.sleep(cfg_input["retry"])



import asyncio
import logging
import sys

clogger = logging.getLogger("dnstap_receiver.console")
tap_logger = logging.getLogger("dnstap_receiver.output.stdout")

from dnstap_receiver.outputs import transform

def checking_conf(cfg):
    """validate the config"""
    clogger.debug("Output handler: stdout")
    return True
    
def setup_logger():
    """setup loggers"""
    logfmt = '%(message)s'
    
    tap_logger.setLevel(logging.INFO)
    tap_logger.propagate = False
    
    lh = logging.StreamHandler(stream=sys.stdout)
    lh.setLevel(logging.INFO)
    lh.setFormatter(logging.Formatter(logfmt))    
    
    tap_logger.addHandler(lh)
    
async def handle(output_cfg, queue, metrics, start_shutdown):
    """stdout output handler"""
    
    # init logger
    setup_logger()
    
    while not start_shutdown.is_set():
        # read item from queue
        try:
            tapmsg = await asyncio.wait_for(queue.get(), timeout=0.5)
        except asyncio.TimeoutError:
            continue
        
        # convert dnstap message
        msg = transform.convert_dnstap(fmt=output_cfg["format"], tapmsg=tapmsg)
        
        # print to stdout
        tap_logger.info(msg.decode())
        
        # all done
        queue.task_done()



import asyncio
import logging
import threading

try:
    import confluent_kafka
    has_kafka = True
except:
    has_kafka = False

clogger = logging.getLogger("dnstap_receiver.console")

from dnstap_receiver.outputs import transform


class Producer:
    def __init__(self, config, topic, start_shutdown):
        self.loop = asyncio.get_event_loop()
        self.producer = confluent_kafka.Producer(config)
        self.topic = topic
        self.start_shutdown = start_shutdown

        self.polling_task = asyncio.create_task(self.polling_task())


    async def polling_task(self):
        while not self.start_shutdown.is_set():
            await asyncio.to_thread(self.producer.poll, 1)

        clogger.debug("Output handler: kafka: performing last flush")
        self.producer.flush()
        clogger.info("Output handler: kafka: polling task stopped")


    def produce(self, value):
        self.producer.produce(self.topic, value)
        self.producer.poll(0)


def checking_conf(cfg):
    """validate the config"""
    clogger.debug("Output handler: kafka")

    valid_conf = True

    if not has_kafka:
        valid_conf = False
        clogger.error("Output handler: kafka: confluent_kafka dependency is missing")

    if cfg["rdkafka-config"]["bootstrap.servers"] is None:
        valid_conf = False
        clogger.error("Output handler: kafka: no bootstrap.servers provided")

    if cfg["topic"] is None:
        valid_conf = False
        clogger.error("Output handler: kafka: no topic provided")

    return valid_conf


async def handle(output_cfg, queue, metrics, start_shutdown):
    start_shutdown_producer = asyncio.Event()
    producer = Producer(output_cfg['rdkafka-config'], output_cfg['topic'], start_shutdown_producer)

    clogger.info("Output handler: kafka: Enabled")
    while not start_shutdown.is_set():
        try:
            tapmsg = await asyncio.wait_for(queue.get(), timeout=0.5)
        except asyncio.TimeoutError:
            continue
        msg = transform.convert_dnstap(fmt=output_cfg["format"], tapmsg=tapmsg)

        producer.produce(msg)
        queue.task_done()

    # tell producer to shut down
    clogger.info("Output handler: kafka: Triggering producer shutdown")
    start_shutdown_producer.set()




import struct

DNS_LEN = 12

unpack_dns = struct.Struct("!6H").unpack

def decode_question(data):
    """minimalist decoder for dns question"""
    buf = data[DNS_LEN:]
    qname = []

    while len(buf):
        length = buf[0]
        if length == 0x00:
            break
        label = buf[1:length+1]
        qname.append(buf[1:length+1])
        buf = buf[length+1:]

    # requires a buffer of 4 bytes 
    # fix issue #35, invalid packet but some abuser sends it
    if len(buf) <= 4:
        qtype = 0 # set to None
    else:
        # decode qtype and qclass 
        qtype, qclass  = struct.unpack('!HH', buf[1:5])    
    return (b".".join(qname)+ b".", qtype) 

def decode_dns(data):
    dns_hdr = unpack_dns(data[:DNS_LEN])
    dns_id = dns_hdr[0]
    dns_rcode = dns_hdr[1] & 15
    dns_qdcount = dns_hdr[2]
    
    return (dns_id, dns_rcode, dns_qdcount)
    
def print_dns(data):
    dns_id, dns_rcode, dns_qdcount = decode_dns(data)
    if dns_qdcount:
        qname, qtype = decode_question(data)


import asyncio
import logging

clogger = logging.getLogger("dnstap_receiver.console")
clogger = logging.getLogger("dnstap_receiver.console")

from dnstap_receiver.outputs import transform

def checking_conf(cfg):
    """validate the config"""
    clogger.debug("Output handler: tcp")
    
    valid_conf = True
    
    if cfg["remote-address"] is None:
        valid_conf = False
        clogger.error("Output handler: no remote address provided")
        
    if cfg["remote-port"] is None:
        valid_conf = False
        clogger.error("Output handler: no port provided")
            
    return valid_conf
    
async def plaintext_tcpclient(output_cfg, queue, start_shutdown):
    host, port = output_cfg["remote-address"], output_cfg["remote-port"]
    clogger.debug("Output handler: connection to %s:%s" % (host,port) )
    reader, tcp_writer = await asyncio.open_connection(host, port)
    clogger.debug("Output handler: connected")
    
    # consume queue
    while not start_shutdown.is_set():
        # read item from queue
        tapmsg = await queue.get()
        
        # convert dnstap message
        msg = transform.convert_dnstap(fmt=output_cfg["format"], tapmsg=tapmsg)
            
        # write & add delimiter
        tcp_writer.write(msg + output_cfg["delimiter"].encode())
        
        # connection lost ? exit and try to reconnect 
        if tcp_writer.transport._conn_lost:
            break
        
        # done continue to next item
        queue.task_done()
        
    # something 
    if not start_shutdown.is_set():
        clogger.error("Output handler: connection lost")

async def handle(output_cfg, queue, metrics, start_shutdown):
    """tcp reconnect"""
    server_address = (output_cfg["remote-address"], output_cfg["remote-port"])
    loop = asyncio.get_event_loop()

    clogger.debug("Output handler: TCP enabled")
    while not start_shutdown.is_set():
        try:
            await plaintext_tcpclient(output_cfg, queue, start_shutdown)
        except ConnectionRefusedError:
            clogger.error('Output handler: connection to tcp server failed!')
        except asyncio.TimeoutError:
            clogger.error('Output handler: connection to tcp server timed out!')
        else:
            clogger.error('Output handler: connection to tcp is closed.')
            
        clogger.debug("'Output handler: retry to connect every %ss" % output_cfg["retry"])
        if not start_shutdown.is_set():
            await asyncio.sleep(output_cfg["retry"])




RDATATYPES = ["-"]*65535
RDATATYPES[0] = "NONE"
RDATATYPES[1] = "A"
RDATATYPES[2] = "NS"
RDATATYPES[3] = "MD"
RDATATYPES[4] = "MF"
RDATATYPES[5] = "CNAME"
RDATATYPES[6] = "SOA"
RDATATYPES[7] = "MB"
RDATATYPES[8] = "MG"
RDATATYPES[9] = "MR"
RDATATYPES[10] = "NULL"
RDATATYPES[11] = "WKS"
RDATATYPES[12] = "PTR"
RDATATYPES[13] = "HINFO"
RDATATYPES[14] = "MINFO"
RDATATYPES[15] = "MX"
RDATATYPES[16] = "TXT"
RDATATYPES[17] = "RP"
RDATATYPES[18] = "AFSDB"
RDATATYPES[19] = "X25"
RDATATYPES[20] = "ISDN"
RDATATYPES[21] = "RT"
RDATATYPES[22] = "NSAP"
RDATATYPES[23] = "NSAP_PTR"
RDATATYPES[24] = "SIG"
RDATATYPES[25] = "KEY"
RDATATYPES[26] = "PX"
RDATATYPES[27] = "GPOS"
RDATATYPES[28] = "AAAA"
RDATATYPES[29] = "LOC"
RDATATYPES[30] = "NXT"
RDATATYPES[33] = "SRV"
RDATATYPES[35] = "NAPTR"
RDATATYPES[36] = "KX"
RDATATYPES[37] = "CERT"
RDATATYPES[38] = "A6"
RDATATYPES[39] = "DNAME"
RDATATYPES[41] = "OPT"
RDATATYPES[42] = "APL"
RDATATYPES[43] = "DS"
RDATATYPES[44] = "SSHFP"
RDATATYPES[45] = "IPSECKEY"
RDATATYPES[46] = "RRSIG"
RDATATYPES[47] = "NSEC"
RDATATYPES[48] = "DNSKEY"
RDATATYPES[49] = "DHCID"
RDATATYPES[50] = "NSEC3"
RDATATYPES[51] = "NSEC3PARAM"
RDATATYPES[52] = "TSLA"
RDATATYPES[53] = "SMIMEA"
RDATATYPES[55] = "HIP"
RDATATYPES[56] = "NINFO"
RDATATYPES[59] = "CDS"
RDATATYPES[60] = "CDNSKEY"
RDATATYPES[61] = "OPENPGPKEY"
RDATATYPES[62] = "CSYNC"
RDATATYPES[64] = "SVCB"
RDATATYPES[65] = "HTTPS"
RDATATYPES[99] = "SPF"
RDATATYPES[103] = "UNSPEC"
RDATATYPES[108] = "EUI48"
RDATATYPES[109] = "EUI64"
RDATATYPES[249] = "TKEY"
RDATATYPES[250] = "TSIG"
RDATATYPES[251] = "IXFR"
RDATATYPES[252] = "AXFR"
RDATATYPES[253] = "MAILB"
RDATATYPES[254] = "MAILA"
RDATATYPES[255] = "ANY"
RDATATYPES[256] = "URI"
RDATATYPES[257] = "CAA"
RDATATYPES[258] = "AVC"
RDATATYPES[259] = "AMTRELAY"
RDATATYPES[32768] = "TA"
RDATATYPES[32769] = "DLV"



RCODES = ["-"]*4095
RCODES[0] = "NOERROR"
RCODES[1] = "FORMERR"
RCODES[2] = "SERVFAIL"
RCODES[3] = "NXDOMAIN"
RCODES[4] = "NOIMP"
RCODES[5] = "REFUSED"
RCODES[6] = "YXDOMAIN"
RCODES[7] = "YXRRSET"
RCODES[8] = "NXRRSET"
RCODES[9] = "NOTAUTH"
RCODES[10] = "NOTZONE"
RCODES[11] = "DSOTYPENI"
RCODES[16] = "BADSIG"
RCODES[17] = "BADKEY"
RCODES[18] = "BADTIME"
RCODES[19] = "BADMODE"
RCODES[20] = "BADNAME"
RCODES[21] = "BADALG"
RCODES[22] = "BADTRUNC"
RCODES[23] = "BADCOOKIE"


import argparse
import logging
import asyncio
import yaml
import signal
import sys

import ssl
import pkgutil
import pathlib
import cachetools

import geoip2.database

# create default logger for the dnstap receiver
clogger = logging.getLogger("dnstap_receiver.console")

# import all inputs
from dnstap_receiver.inputs import input_socket
from dnstap_receiver.inputs import input_sniffer
from dnstap_receiver.inputs import input_tcpclient

# import all outputs
from dnstap_receiver.outputs import output_stdout
from dnstap_receiver.outputs import output_file
from dnstap_receiver.outputs import output_syslog
from dnstap_receiver.outputs import output_tcp
from dnstap_receiver.outputs import output_metrics
from dnstap_receiver.outputs import output_dnstap
from dnstap_receiver.outputs import output_kafka
from dnstap_receiver.outputs import output_pgsql

from dnstap_receiver import api_server
from dnstap_receiver import statistics

DFLT_LISTEN_IP = "0.0.0.0"
DFLT_LISTEN_PORT = 6000

# command line arguments definition
parser = argparse.ArgumentParser()
parser.add_argument("-l", 
                    help="IP of the dnsptap server to receive dnstap payloads (default: %(default)r)",
                    default=DFLT_LISTEN_IP)
parser.add_argument("-p", type=int,
                    help="Port the dnstap receiver is listening on (default: %(default)r)",
                    default=DFLT_LISTEN_PORT)               
parser.add_argument("-u", help="read dnstap payloads from unix socket")
parser.add_argument('-v', action='store_true', help="verbose mode")   
parser.add_argument("-c", help="external config file")   

# get event loop
loop = asyncio.get_event_loop()
shutdown_task = None

def merge_cfg(u, o):
    """merge config"""
    for k,v in u.items():
        if k in o:
            if isinstance(v, dict):
                merge_cfg(u=v,o=o[k])
            else:
                o[k] = v

def load_yaml(f):
    """load yaml file"""
    try:
        cfg =  yaml.safe_load(f) 
    except FileNotFoundError:
        print("default config file not found")
        sys.exit(1)
    except yaml.parser.ParserError:
        print("invalid default yaml config file")
        sys.exit(1)
    return cfg 
    
def setup_config(args):
    """load default config and update it with arguments if provided"""
    # Set the default configuration file
    f = pkgutil.get_data(__package__, 'dnstap.conf')
    cfg = load_yaml(f)

    # Overwrites then with the external file ?    
    if args.c:
        cfg_ext = load_yaml(open(args.c, 'r'))
        merge_cfg(u=cfg_ext,o=cfg)

    # Or searches for a file named dnstap.conf in /etc/dnstap_receiver/       
    else:
        etc_conf = "/etc/dnstap_receiver/dnstap.conf"
        f = pathlib.Path(etc_conf)
        if f.exists():
            cfg_etc = load_yaml(open(etc_conf, 'r'))
            merge_cfg(u=cfg_etc,o=cfg)
            
    # update default config with command line arguments
    if args.v:
        cfg["trace"]["verbose"] = args.v    
    if args.u is not None:
        cfg["input"]["unix-socket"]["enable"] = True
        cfg["input"]["unix-socket"]["path"] = args.u
    if args.l != DFLT_LISTEN_IP:
        cfg["input"]["tcp-socket"]["local-address"] = args.l
    if args.l != DFLT_LISTEN_PORT:
        cfg["input"]["tcp-socket"]["local-port"] = args.p

    return cfg
    
def setup_logger(cfg):
    """setup main logger"""

    loglevel = logging.DEBUG if cfg["verbose"] else logging.INFO
    logfmt = '%(asctime)s %(levelname)s %(message)s'
    
    clogger.setLevel(loglevel)
    clogger.propagate = False
    
    if cfg["file"] is None:
        lh = logging.StreamHandler(stream=sys.stdout )
    else:
        lh = logging.FileHandler(cfg["file"])
    lh.setLevel(loglevel)
    lh.setFormatter(logging.Formatter(logfmt))    
    
    clogger.addHandler(lh)
    
def setup_outputs(cfg, stats, start_shutdown):
    """setup outputs"""
    conf = cfg["output"]

    queues_list = []
    if conf["syslog"]["enable"]:
        if not output_syslog.checking_conf(cfg=conf["syslog"]): return
        queue_syslog = asyncio.Queue()
        queues_list.append(queue_syslog)
        loop.create_task(output_syslog.handle(conf["syslog"], queue_syslog, stats, start_shutdown))

    if conf["tcp-socket"]["enable"]:
        if not output_tcp.checking_conf(cfg=conf["tcp-socket"]): return
        queue_tcpsocket = asyncio.Queue()
        queues_list.append(queue_tcpsocket)
        loop.create_task(output_tcp.handle(conf["tcp-socket"], queue_tcpsocket, stats, start_shutdown))
                                               
    if conf["file"]["enable"]:
        if not output_file.checking_conf(cfg=conf["file"]): return
        queue_file = asyncio.Queue()
        queues_list.append(queue_file)
        loop.create_task(output_file.handle(conf["file"], queue_file, stats, start_shutdown))
                                              
    if conf["stdout"]["enable"]:
        if not output_stdout.checking_conf(cfg=conf["stdout"]): return
        queue_stdout = asyncio.Queue()
        queues_list.append(queue_stdout)
        loop.create_task(output_stdout.handle(conf["stdout"], queue_stdout, stats, start_shutdown))

    if conf["metrics"]["enable"]:
        if not output_metrics.checking_conf(cfg=conf["metrics"]): return
        queue_metrics = asyncio.Queue()
        queues_list.append(queue_metrics)
        loop.create_task(output_metrics.handle(conf["metrics"], queue_metrics, stats, start_shutdown))

    if conf["dnstap"]["enable"]:
        if not output_dnstap.checking_conf(cfg=conf["dnstap"]): return
        queue_dnstap = asyncio.Queue()
        queues_list.append(queue_dnstap)
        loop.create_task(output_dnstap.handle(conf["dnstap"], queue_dnstap, stats, start_shutdown))

    if conf["kafka"]["enable"]:
        if not output_kafka.checking_conf(cfg=conf["kafka"]): return
        queue_kafka = asyncio.Queue()
        queues_list.append(queue_kafka)
        loop.create_task(output_kafka.handle(conf["kafka"], queue_kafka, stats, start_shutdown))

    if conf["pgsql"]["enable"]:
        if not output_pgsql.checking_conf(cfg=conf["pgsql"]): return
        queue_pgsql = asyncio.Queue()
        queues_list.append(queue_pgsql)
        loop.create_task(output_pgsql.handle(conf["pgsql"], queue_pgsql, stats, start_shutdown))

    return queues_list
    
def setup_inputs(cfg, queues_outputs, stats, geoip_reader, start_shutdown):
    """setup inputs"""
    cache = cachetools.TTLCache(maxsize=1000000, ttl=60)
    cfg_input = cfg["input"]
    
    # asynchronous unix 
    if cfg_input["unix-socket"]["enable"]:
        loop.create_task(input_socket.start_unixsocket(cfg, queues_outputs, stats, geoip_reader, cache, start_shutdown))
        
    # sniffer
    elif cfg_input["sniffer"]["enable"]:
        queue_sniffer = asyncio.Queue()
        loop.create_task(input_sniffer.watch_buffer(cfg_input["sniffer"], queue_sniffer, queues_outputs, stats, cache, start_shutdown))
        loop.run_in_executor(None, input_sniffer.start_input, cfg_input["sniffer"], queue_sniffer, start_shutdown)
    
    # tcp client input
    elif cfg_input["tcp-client"]["enable"]:
        loop.create_task(input_socket.start_tcpclient(cfg, queues_outputs, stats, geoip_reader, cache, start_shutdown))
        
    # default one tcp socket
    else:
        loop.create_task(input_socket.start_tcpsocket(cfg, queues_outputs, stats, geoip_reader, cache, start_shutdown))

def setup_webserver(cfg, stats):
    """setup web api"""
    if not cfg["web-api"]["enable"]: return

    svr = api_server.create_server(loop, cfg=cfg["web-api"], stats=stats, cfg_stats=cfg["statistics"])
    if svr is None: return
        
    loop.create_task( svr)
    
def setup_geoip(cfg):
    if not cfg["enable"]: return None
    if cfg["city-database"] is None: return None
    
    reader = None
    try:
        reader = geoip2.database.Reader(cfg["city-database"])
    except Exception as e:
        clogger.error("geoip setup: %s" % e)
        
    return reader

def start_shutdown_task(signal, loop, start_shutdown):
    global shutdown_task
    if not shutdown_task:
        shutdown_task = asyncio.create_task(
            shutdown(signal, loop, start_shutdown)
        )

async def shutdown(signal, loop, start_shutdown):
    """perform graceful shutdown"""

    clogger.info("starting shutting down process")
    start_shutdown.set()

    current_task = asyncio.current_task()
    tasks = [
        task for task in asyncio.all_tasks()
        if task is not current_task
    ]

    clogger.info("waiting for all tasks to exit")
    await asyncio.gather(*tasks, return_exceptions=True)

    clogger.debug("all tasks have exited, stopping loop")
    loop.stop()


def start_receiver():
    """start dnstap receiver"""
    # Handle command-line arguments.
    args = parser.parse_args()

    # setup config
    cfg = setup_config(args=args)
  
    # setup logging
    setup_logger(cfg=cfg["trace"])

    # setup geoip if enabled 
    geoip_reader = setup_geoip(cfg=cfg["geoip"])
    
    # prepare shutdown handling
    start_shutdown = asyncio.Event()
    for sig in (signal.SIGHUP, signal.SIGTERM, signal.SIGINT):
        loop.add_signal_handler(sig, lambda: start_shutdown_task(sig, loop, start_shutdown))

    # add debug message if external config is used
    if args.c: clogger.debug("External config file loaded")
    
    # start receiver
    clogger.debug("Start receiver...")
    stats = statistics.Statistics(cfg=cfg["statistics"])
    loop.create_task(statistics.watcher(stats, start_shutdown))
    
    # prepare outputs
    queues_outputs = setup_outputs(cfg, stats, start_shutdown)

    # prepare inputs
    setup_inputs(cfg, queues_outputs, stats, geoip_reader, start_shutdown)

    # start the http api
    setup_webserver(cfg, stats)
        
    # run event loop 
    try:
       loop.run_forever()
    finally:
        clogger.info("exiting, please wait..")
        loop.close()
        clogger.debug("shut down eventloop")

    # close geoip
    if geoip_reader is not None: geoip_reader.close()



import asyncio
import logging
import ssl
import ipaddress
import fstrm
import dnstap_pb

from dnstap_receiver.inputs import dnstap_decoder 

clogger = logging.getLogger("dnstap_receiver.console")


async def cb_onconnect(reader, writer, cfg, cfg_input, queues_list, stats, geoip_reader, cache, start_shutdown):
    """callback when a connection is established"""
    # get peer name
    peername = writer.get_extra_info('peername')
    if not len(peername):
        peername = "(unix-socket)"
    clogger.debug(f"Input handler: new connection from {peername}")

    # access control list check
    if "access-control-list" in cfg_input:
        if len(writer.get_extra_info('peername')):
            acls_network = []
            for a in cfg_input["access-control-list"]:
                acls_network.append(ipaddress.ip_network(a))
                
            acl_allow = False
            for acl in acls_network:
                if ipaddress.ip_address(peername[0]) in acl:
                    acl_allow = True
            
            if not acl_allow:
                writer.close()
                clogger.debug("Input handler: checking acl refused")
                return
            
            clogger.debug("Input handler: checking acl allowed")
        
    # prepare frame streams decoder
    content_type = b"protobuf:dnstap.Dnstap"
    fstrm_handler = fstrm.FstrmCodec()
    loop = asyncio.get_event_loop()
    dnstap_protobuf = dnstap_pb.Dnstap()

    try: 
        # syntax only works with python 3.8
        # while data := await reader.read(fstrm_handler.pending_nb_bytes()) 
        running = True
        while running:
            # read bytes
            shutdown_wait_task = asyncio.create_task(start_shutdown.wait())
            read_task = asyncio.create_task(reader.read(fstrm_handler.pending_nb_bytes()))
            done, pending = await asyncio.wait(
                [shutdown_wait_task, read_task],
                return_when=asyncio.FIRST_COMPLETED,
            )

            if shutdown_wait_task in done:
                read_task.cancel()
                return
            else:
                shutdown_wait_task.cancel()
                data = await read_task

            if not len(data):
                running = False
                break
                
            # append data to the buffer
            fstrm_handler.append(data=data)
            
            # process the buffer, check if we have received a complete frame ?
            if fstrm_handler.process():
                # Ok, the frame is complete so let's decode it
                ctrl, ct, payload  = fstrm_handler.decode()

                # handle the DATA frame
                if ctrl == fstrm.FSTRM_DATA_FRAME:
                    await dnstap_decoder.cb_ondnstap(dnstap_protobuf, payload, cfg, queues_list, stats, geoip_reader, cache)
                    
                # handle the control frame READY
                if ctrl == fstrm.FSTRM_CONTROL_READY:
                    clogger.debug(f"Input handler: control ready received from {peername}")
                    if content_type not in ct:
                        raise Exception("content type error: %s" % ct)
                        
                    # todo, checking content type
                    ctrl_accept = fstrm_handler.encode(ctrl=fstrm.FSTRM_CONTROL_ACCEPT, ct=[content_type])
                    # respond with accept only if the content type is dnstap
                    writer.write(ctrl_accept)
                    await writer.drain()
                    clogger.debug(f"Input handler: sending control accept to {peername}")
                    
                # handle the control frame READY
                if ctrl == fstrm.FSTRM_CONTROL_START:
                    clogger.debug(f"Input handler: control start received from {peername}")
   
                # handle the control frame STOP
                if ctrl == fstrm.FSTRM_CONTROL_STOP:
                    clogger.debug(f"Input handler: control stop received from {peername}")
                    fstrm_handler.reset()
                    
                    # send finish control
                    ctrl_finish = fstrm_handler.encode(ctrl=fstrm.FSTRM_CONTROL_FINISH)
                    writer.write(ctrl_finish)
                    await writer.drain()
                    
                    clogger.debug(f"Input handler: sending control finish to {peername}")
                    
    except ConnectionError as e:
        clogger.error(f'Input handler: {peername} - %s' % e)
    except asyncio.CancelledError:
        clogger.debug(f'Input handler: {peername} - closing connection.')
        writer.close()
        await writer.wait_closed()
    except asyncio.IncompleteReadError:
        clogger.debug(f'Input handler: {peername} - disconnected')
    finally:
        clogger.debug(f'Input handler: {peername} - closed')

def start_tcpsocket(cfg, queues_list, stats, geoip_reader, cache, start_shutdown):
    clogger.debug("Input handler: tcp socket")
    loop = asyncio.get_event_loop()
    cfg_input = cfg["input"]["tcp-socket"]
    
    # define callback on new connection
    cb_lambda = lambda r, w: cb_onconnect(r, w, cfg, cfg_input, queues_list, stats, geoip_reader, cache, start_shutdown)
    
    ssl_context = None
    if cfg_input["tls-support"]:
        ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        ssl_context.load_cert_chain(certfile=cfg_input["tls-server-cert"], keyfile=cfg_input["tls-server-key"])
        clogger.debug("Input handler - tls support enabled")
        
    clogger.debug("Input handler: listening on %s:%s" % (cfg_input["local-address"],cfg_input["local-port"])), 
    server = asyncio.start_server(cb_lambda, cfg_input["local-address"],cfg_input["local-port"],
                                  ssl=ssl_context, loop=loop)
    return server
    
def start_unixsocket(cfg, queues_list, stats, geoip_reader, cache, start_shutdown):
    clogger.debug("Input handler: unix socket")
    cfg_input = cfg["input"]["unix-socket"]
    
    loop = asyncio.get_event_loop()
    
    # define callback on new connection
    cb_lambda = lambda r, w: cb_onconnect(r, w, cfg, cfg_input, queues_list, stats, geoip_reader, cache, start_shutdown)
    
    # asynchronous unix socket
    clogger.debug("Input handler: listening on %s" % cfg_input["path"])
    server = asyncio.start_unix_server(cb_lambda, path=cfg_input["path"], loop=loop)
                                                  
    return server



#!/usr/bin/python

import setuptools
       
with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()
    
KEYWORDS = ('dnstap receiver client json yaml text')

setuptools.setup(
    name="dnstap_receiver",
    version="4.4.0",
    author="Denis MACHARD",
    author_email="d.machard@gmail.com",
    description="Python Dnstap receiver",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/dmachard/dnstap_receiver",
    packages=['dnstap_receiver'],
    include_package_data=True,
    platforms='any',
    keywords=KEYWORDS,
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries",
    ],
    entry_points={'console_scripts': ['dnstap_receiver = dnstap_receiver.receiver:start_receiver']},
    install_requires=[
        "dnstap_pb",
        "pyyaml",
        "aiohttp",
        "geoip2",
        "tlds",
        "cachetools",
        "fstrm"
    ],
    extras_require={
        "kafka": ["confluent-kafka"],
        "pgsql": ["asyncpg"],
    }
)


'''
This file contains the default functions of pgsql_init and pgsql_main.
If you feel like replace them, 
  1) copy this file and edit it as you like (but keep function names),
  2) then designate it in the `userfuncefile` in the configuration.
     (See outputs: pgsql: section in ../dnstap.conf)
'''

import logging
clogger = logging.getLogger("dnstap_receiver.console")

async def pgsql_init(conn):
    '''
    pgsql_init is a function which is executed once just after
    creation of asyncpg connection pool (nearly equals to every time 
    when the dnstap_receiver being started).
    It is expected to do something like "CREATE TABLE IF NOT EXISTS..." here.

    `conn` is a connection to PostgreSQL server acquired from pool.
    '''
    clogger.info("pgsql_init: createing table if not exists.")
    return await conn.execute("""
        CREATE TABLE IF NOT EXISTS dnstap_receiver (
            message     TEXT        -- "AUTH_QUERY"
           ,type        TEXT        -- "query"
           ,timestamp   TIMESTAMPTZ -- "1625636652.113565"
           ,query_ip    TEXT        -- "192.0.2.100"
           ,response_ip TEXT        -- "203.0.113.200"
           ,qname       TEXT        -- "www.example.com."
           ,rrtype      TEXT        -- "A"
           ,rcode       TEXT        -- "NOERROR"
        )
    """)

async def pgsql_main(tapmsg, conn):
    '''
    pgsql_main is a function which is executed on each dnstap data delivered.
    It is expected to do something like "INSERT INTO..." here.

    `conn` is a connection to PostgreSQL server acquired from pool.
    `tapmsg` is a dict that contains dnstap data delivered.
    '''
    clogger.info("pgsql_main: inserting data.")
    return await conn.execute("""
        INSERT INTO dnstap_receiver VALUES
            ($1, $2, to_timestamp($3), $4, $5, $6, $7, $8)
        """,
            tapmsg['message'], tapmsg['type']
           ,tapmsg['timestamp'], tapmsg['query-ip']
           ,tapmsg['response-ip'], tapmsg['qname']
           ,tapmsg['rrtype'], tapmsg['rcode']
    )



import asyncio
import logging

from os.path import abspath,realpath,expanduser,expandvars
from importlib.util import spec_from_file_location, module_from_spec

try:
    import asyncpg
    has_pgsql = True
except:
    has_pgsql = False

clogger = logging.getLogger("dnstap_receiver.console")

from dnstap_receiver.outputs import transform


def checking_conf(cfg):
    """validate the config"""
    clogger.debug("Output handler: pgsql")

    valid_conf = True

    if not has_pgsql:
        valid_conf = False
        clogger.error("Output handler: pgsql: asyncpg dependency is missing")

    if cfg["dsn"] is None:
        valid_conf = False
        clogger.error("Output handler: no dsn provided")

    return valid_conf

async def plaintext_pgclient(output_cfg, queue, start_shutdown):
    dsn = output_cfg["dsn"]
    clogger.debug("Output handler: connection to %s" % (dsn,))

    passfile = output_cfg["passfile"]
    min_size = output_cfg["min_size"]
    max_size = output_cfg["max_size"]
    busy_wait = float(output_cfg["busy_wait"])
    userfuncfile = output_cfg["userfuncfile"]

    # importing functions to handle PostgreSQL.
    # pgsql_init shall be executed once just after connection pool
    # to PostgreSQL. Ususally it should contain "CREATE TABLE IF NOT
    # EXISTS..."
    # pgsql_main shall be executed on receiving every DNS queries.
    # Usually it should be "INSERT INTO..."
    # dnstap_receiver has default functions to fall back to, or 
    # user can define his/her own function in the 'userfuncfile'.
    # For example,
    # $ cp output_pgsql_userfunc.py output_pgsql_myfunc.py
    # $ vi output_pgsql_myfunc.py
    # and make 'userfuncfile: /path/to/output_pgsql_myfunc.py' in dnstap.conf
    if userfuncfile is None:
        clogger.debug(f"Output handler: pgsql: loading default userfuncfile.")
        from .output_pgsql_userfunc import pgsql_init, pgsql_main
    else:
        try:
            userfuncfile = abspath(realpath(expandvars(expanduser(userfuncfile))))
            # Should check process euid == file owner ?

            spec = spec_from_file_location('userfunc', userfuncfile)
            userfunc = module_from_spec(spec)
            spec.loader.exec_module(userfunc)
            pgsql_init = userfunc.pgsql_init
            pgsql_main = userfunc.pgsql_main
            clogger.debug(f"Output handler: pgsql: loaded userfunc in {userfuncfile}.")
        except:
            clogger.info("Output handler: pgsql faild to load userfunc. fallback to default.")
            from .output_pgsql_userfunc import pgsql_init, pgsql_main

    # create connection pool to PostgreSQL server.
    async with asyncpg.create_pool(dsn=dsn, passfile=passfile, min_size=min_size, max_size=max_size, timeout=15) as pool:
        clogger.debug("Output handler: pgsql connected")

        # acquire a connection and execute pgsql_init()
        # such as "CREATE TABLE IF NOT EXISTS..."
        async with pool.acquire() as conn:
            async with conn.transaction():
                await pgsql_init(conn)

        # consume queue
        while not start_shutdown.is_set():
            #clogger.debug(f'Output handler: pgsql receiving tapmsg from queue.')
            # 'tapmsg = await queue.get()' will block start_shutdown_task
            # to gracefully shutdown dnstap_receiver itself.
            # 'queue.get_nowait()' won't block but introduces
            # busy-wait-loop instead. which do yo like?
            try:
                tapmsg = queue.get_nowait()
            except asyncio.QueueEmpty as e:
                if start_shutdown.is_set():
                    clogger.debug('Output handler: pgsql shutting down. ')
                    break
                else:
                    await asyncio.sleep(busy_wait)
                    continue
            else:
                clogger.debug(f'Output handler: pgsql received tapmsg: {tapmsg}.')

            # acquire a connection and send 'INSERT...' to PostgreSQL server.
            async with pool.acquire() as conn:
                async with conn.transaction():
                    await pgsql_main(tapmsg, conn)
                    clogger.debug('Output handler: pgsql INSERT dispached.')
    
            # done continue to next item
            queue.task_done()

        clogger.debug(f'Output handler: pgsql closing pool.')

    # something 
    if not start_shutdown.is_set():
        clogger.error("Output handler: pgclient connection lost")

async def handle(output_cfg, queue, metrics, start_shutdown):
    """pgsql reconnect"""
    loop = asyncio.get_event_loop() # do we need this?

    clogger.debug("Output handler: PostgreSQL enabled")

    while not start_shutdown.is_set():
        try:
            await plaintext_pgclient(output_cfg, queue, start_shutdown)
        except ConnectionRefusedError:
            clogger.error('Output handler: connection to pgsql server failed!')
        except asyncio.TimeoutError:
            clogger.error('Output handler: connection to pgsql server timed out!')
        else:
            clogger.error('Output handler: connection to pgsql is closed.')

        if not start_shutdown.is_set():
            clogger.debug("'Output handler: retry to connect every %ss" % output_cfg["retry"])
            await asyncio.sleep(output_cfg["retry"])
